import os, sys, time, random
from qgis.core import *
from qgis.gui import *
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from utility.validateUi import ValidateUi
from utility.validators import StructureValidator, ShapefileValidator
from utility.shp_errors import ErrorShapefile
from datetime import datetime

class ProgressUi(QDialog):

	_HEADERS = ["Step", "Tipo controllo", "Stato", "Validazione"]
	_INITVALUES = {0: ["1/5", "2/5", "3/5", "4/5", "5/5"],
				   1: ["Struttura progetto", "Proprieta' e schema dei files", "Domini delle features",
								   "Validita' geometrica", "Topologia"],
				   2: ["In attesa..."] * 5,
				   3: [""] * 5}

	def __init__(self, validateParameters, parent=None):
		super(ProgressUi, self).__init__(parent)
		self.setAttribute(Qt.WA_DeleteOnClose)

		self.validateParameters = validateParameters

		self.browser = QTextBrowser()

		self.table = QTableWidget(5, 4)
		self.table.setHorizontalHeaderLabels(ProgressUi._HEADERS)
		self.table.setMaximumHeight(140)
		self.table.resizeRowsToContents()
		self.table.setWordWrap(True)

		vHeader = self.table.verticalHeader()
		vHeader.hide()

		self.table.setColumnWidth(0, 50)
		self.table.setColumnWidth(1, 250)
		self.table.setColumnWidth(2, 150)
		self.table.setColumnWidth(3, 124)


		self.runButton = QPushButton("Avvia")
		self.saveButton = QPushButton("Salva")
		self.closeButton = QPushButton("Chiudi")

		progressLayout = QVBoxLayout()
		progressLayout.addWidget(self.table)
		progressLayout.addWidget(self.browser)

		buttonLayout = QHBoxLayout()
		buttonLayout.addStretch()
		buttonLayout.addWidget(self.runButton)
		buttonLayout.addWidget(self.saveButton)
		buttonLayout.addWidget(self.closeButton)

		mainLayout = QVBoxLayout()
		mainLayout.addLayout(progressLayout)
		mainLayout.addLayout(buttonLayout)

		self.setLayout(mainLayout)
		self.setFixedSize(600, 500)
		self.setWindowTitle("Validazione")

		self.connect(self.runButton, SIGNAL("clicked()"), self.run)
		self.connect(self.saveButton, SIGNAL("clicked()"), self.saveReport)
		self.connect(self.closeButton, SIGNAL("clicked()"), self, SLOT("reject()"))

		self.addRecord()

		self.printParams()

	def saveReport(self):
		t = datetime.today()
		today = t.strftime('%Y%m%d%H%M') + "_"
		with open(os.path.join(self.validateParameters["path"], "Progetti", "reports", "Report_{0}.txt".format(today)), 'w') as f:
			f.write(str(self.browser.toPlainText()))

		msgBox = QMessageBox()
		msgBox.setText("Report salvato correttamente")
		msgBox.exec_()

	def addRecord(self):
		for row in range(len(ProgressUi._INITVALUES[0])):
			for col in range(len(ProgressUi._HEADERS)):
				newitem = QTableWidgetItem(ProgressUi._INITVALUES[col][row])
				if col == 3:
					font = newitem.font()
					font.setBold(True)
					newitem.setTextColor(QColor("darkGreen"))
					newitem.setFont(font)
				self.table.setItem(row, col, newitem)

	def setRecord(self, step, status, validation=None):
		if status == 2:
			statusName = "In attesa..."
		elif status == 1:
			statusName = "In corso..."
		elif status == 0:
			statusName = "Completato"

		statusItem = QTableWidgetItem(statusName)

		self.table.setItem(step - 1, 2, statusItem)
		QApplication.processEvents()

		if validation is not None:
			if validation:
				validationItem = QTableWidgetItem("Passed")
				validationItem.setTextColor(QColor("darkGreen"))
			else:
				validationItem = QTableWidgetItem("Not Passed")
				validationItem.setTextColor(QColor("darkRed"))

			font = validationItem.font()
			font.setBold(True)
			validationItem.setFont(font)

			self.table.setItem(step - 1, 3, validationItem)
			QApplication.processEvents()

	def printParams(self):
		self.browser.append("<br />")
		self.browser.append("Percorso di analisi: <b><i>%s</i></b>" % self.validateParameters["path"])
		self.browser.append("Formato di archiviazione: <b><i>%s</i></b>" % ("Shapefile" if self.validateParameters["type"] == "shp" else "Geodatabase"))
		self.browser.append("Versione di archiviazione: <b><i>%d</i></b>" % self.validateParameters["version"])
		self.browser.append("<br />")

		self.browser.append("Premi il tasto <b>\"Avvia\"</b> per iniziare la procedura di validazione<br />")

	"""
	def openLastWindow(self):
		self.reject()
		form = ValidateUi(self.validateParameters)
		if form.exec_():
			newinstance = ProgressUi(self.validateParameters)
			newinstance.exec_()


	def test(self):
		self.runButton.setEnabled(False)
		self.closeButton.setEnabled(False)
		for step in range(1, 6):
			self.setRecord(step, 1)
			QApplication.processEvents()
			time.sleep(5)
			self.setRecord(step, 0, bool(random.randint(0, 1)))
			QApplication.processEvents()
		self.closeButton.setEnabled(True)
	"""

	def run(self):
		self.runButton.setEnabled(False)
		self.saveButton.setEnabled(False)
		self.closeButton.setEnabled(False)

		mainCheck = {}

		# Inizio controllo Struttura
		self.browser.append("<b><i>[1/5]</i> - Inizio controllo della struttura del progetto...</b>")
		self.browser.append("<br />")

		ctrlStructure = StructureValidator(self.validateParameters["path"],
										   self.validateParameters["version"],
										   self.validateParameters["type"])

		QApplication.processEvents()
		self.setRecord(1, 1)

		checkDirs = []
		firstFileList = []
		for singleCheck in ctrlStructure.check_data_structure():
			checkDirs.append(singleCheck["check"])
			if not singleCheck["check"]:
				msg = "<li>File %s non presente</li>" % singleCheck["path"]
				self.browser.append(msg)
			else:
				if singleCheck["type"] == "shp":
					firstFileList.append(singleCheck["path"])
		if all(checkDirs):
			self.browser.append("Nessun errore riscontrato")

		self.browser.append("<br />")
		self.browser.append("<b><i>[1/5]</i> - Controllo completato</b>")
		self.browser.append("<br />")

		mainCheck[1] = all(checkDirs)

		QApplication.processEvents()
		self.setRecord(1, 0, True) if mainCheck[1] else self.setRecord(1, 0, False)

		# Inizio controllo proprieta files
		self.browser.append("<b><i>[2/5]</i> - Inizio controllo delle proprieta e dello schema dei files...</b>")
		#self.browser.append("<br />")

		QApplication.processEvents()
		self.setRecord(2, 1)

		checkFiles = []
		secondStepFileList = []
		thirdStepFileList = []
		for singleFile in firstFileList:
			shpDir, shp = os.path.split(singleFile)
			shpValidator = ShapefileValidator(singleFile)
			checkSingleFile = {"check": True, "path": singleFile, "feats": True}

			self.browser.append("<br />Controllo del file <b>%s</b>" % os.path.join(os.path.basename(shpDir), shp))

			if not shpValidator.isValid:
				self.browser.append("  - File non valido. Potrebbe essere danneggiato")
				checkSingleFile["check"] = False
				continue

			if not shpValidator.hasValidCrs:
				self.browser.append("  - Sistema di riferimento non valido (atteso EPSG:32633)")
				checkSingleFile["check"] = False

			if not shpValidator.hasFeatures:
				self.browser.append("  - Il file non contiene features")
				checkSingleFile["feats"] = False

			#self.browser.append("<br />")
			self.browser.append("<b>  + Controllo dello schema...</b>")
			checkFields = []
			validFields = []
			for field in shpValidator.validateFields():
				self.browser.append("    - Controllo del campo <b>%s</b>" % str(field["field"]))
				if field.has_key("minField"):
					self.browser.append("      ATTENZIONE - IL CAMPO NON E' PRESENTE")
				else:                    
					if field["check"]:
						self.browser.append("      Nessun errore riscontrato")
					else:
						self.browser.append("      Errore nelle seguenti proprieta:")
						for key, value in field["properties"].items():
							if not value:
								self.browser.append("        - %s [Valore atteso: %s]" % (key, str(field["expected"][key])))
								
				checkFields.append(field["check"])
				
				"""
				if field["check"]:
					validFields.append(field["properties"]["name"])
				"""

			if not all(checkFields):
				checkSingleFile["check"] = False

			checkFiles.append(checkSingleFile["check"])
			if checkSingleFile["check"]:
				secondStepFileList.append(checkSingleFile)
			thirdStepFileList.append(checkSingleFile)

		self.browser.append("<br />")
		self.browser.append("<b><i>[2/5]</i> - Controllo completato</b>")
		self.browser.append("<br />")

		mainCheck[2] = all(checkFiles)

		QApplication.processEvents()
		self.setRecord(2, 0, True) if mainCheck[2] else self.setRecord(2, 0, False)

		# Inizio controllo domini
		self.browser.append("<b><i>[3/5]</i> - Inizio controllo dei domini...</b>")
		#self.browser.append("<br />")

		QApplication.processEvents()
		self.setRecord(3, 1)

		checkDomain = []
		for singleFile in secondStepFileList:
			shpDir, shp = os.path.split(singleFile["path"])
			if singleFile["feats"]:
				shpValidator = ShapefileValidator(singleFile["path"])
				checkSingleFile = {"check": True, "path": singleFile}

				self.browser.append("<br />Controllo del file <b>%s</b>" % os.path.join(os.path.basename(shpDir), shp))
				hasDomain = False
				for field in shpValidator.validateAttributes():
					hasDomain = True
					checkDomain.append(field["check"])
					self.browser.append("Controllo dominio per il campo <b>%s</b>" % field["field"])
					if not field["check"]:
						self.browser.append("Errore di dominio alle features con id: %s" % str(field["featuresId"]))
					else:
						self.browser.append("Nessun errore riscontrato")
				if not hasDomain:
					self.browser.append("Il file non contiene campi con dominio")

		self.browser.append("<br />")
		self.browser.append("<b><i>[3/5]</i> - Controllo completato</b>")
		self.browser.append("<br />")

		mainCheck[3] = all(checkDomain)

		QApplication.processEvents()
		self.setRecord(3, 0, True) if mainCheck[3] else self.setRecord(3, 0, False)


		# Inizio controllo validita geometrica
		self.browser.append("<b><i>[4/5]</i> - Inizio controllo della validita' geometrica...</b>")
		#self.browser.append("<br />")

		QApplication.processEvents()
		self.setRecord(4, 1)

		pointGeomErrors = []
		checkGeom = []
		for singleFile in thirdStepFileList:
			shpDir, shp = os.path.split(singleFile["path"])
			if singleFile["feats"]:
				shpValidator = ShapefileValidator(singleFile["path"])
				checkSingleFileGeom = []

				self.browser.append("<br />Controllo del file <b>%s</b>" % os.path.join(os.path.basename(shpDir), shp))

				for geom in shpValidator.validateGeometries():
					if not geom["check"]:
						if type(geom["errors"]) == tuple:
							self.browser.append("%s" % str(geom["errors"][0]))
						else:
							for what,where in geom["errors"]:
								#if type(error) == str:
								#	self.browser.append("%s" % str(error))
								#	continue
								self.browser.append("%s" % what)
								pointGeomErrors.append((where, what))
					checkSingleFileGeom.append(geom["check"])
					checkGeom.append(geom["check"])
				if all(checkSingleFileGeom):
					self.browser.append("Nessun errore riscontrato")

		self.browser.append("<br />")
		self.browser.append("<b><i>[4/5]</i> - Controllo completato</b>")
		self.browser.append("<br />")

		mainCheck[4] = all(checkGeom)

		QApplication.processEvents()
		self.setRecord(4, 0, True) if mainCheck[4] else self.setRecord(4, 0, False)


		# Inizio controllo validita topologica
		self.browser.append("<b><i>[5/5]</i> - Inizio controllo della validita' topologica...</b>")
		#self.browser.append("<br />")

		QApplication.processEvents()
		self.setRecord(5, 1)

		topoGeomErrors = []
		topoErrors = []
		for singleFile in thirdStepFileList:
			shpDir, shp = os.path.split(singleFile["path"])
			if singleFile["feats"]:
				shpValidator = ShapefileValidator(singleFile["path"])
				checkSingleFileGeom = []

				self.browser.append("<br />Controllo del file <b>%s</b>" % os.path.join(os.path.basename(shpDir), shp))

				errorCount = 0
				for error in shpValidator.validateTopology():
					if not error["check"]:
						topoGeomErrors.append((error["intersection"], "Self-Intersection" + " -- " + error["info"]))
						errorCount += 1
					topoErrors.append(error["check"])
					checkSingleFileGeom.append(error["check"])
				if all(checkSingleFileGeom):
					self.browser.append("Nessun errore riscontrato")
				else:
					self.browser.append("Riscontrati %d errori di autointersezione" % errorCount)

		msDirs = []
		for singleDir in os.listdir(self.validateParameters["path"]):
			if singleDir.lower().startswith('ms'):
				msDirs.append(os.path.join(self.validateParameters["path"], singleDir))

		#topoGeomErrors = []
		for singleDir in msDirs:
			stabFile = None
			instabFile = None
			for singleFile in os.listdir(singleDir):
				if singleFile.lower() == "stab.shp":
					stabFile = os.path.join(singleDir, singleFile)
				if singleFile.lower() == "instab.shp":
					instabFile = os.path.join(singleDir, singleFile)
			if stabFile and instabFile:
				stabValidator = ShapefileValidator(stabFile)
				instabValidator = ShapefileValidator(instabFile)
				if stabValidator.hasFeatures and instabValidator.hasFeatures:
					checkSingleFileGeom = []
					self.browser.append("<br />Controllo topologico tra il file <b>%s</b> ed il file <b>%s</b>" %
					(os.path.join(os.path.basename(singleDir), os.path.basename(stabFile)),
					os.path.join(os.path.basename(singleDir), os.path.basename(instabFile))))

					errorCount = 0
					for error in stabValidator.validateTopology(instabValidator):
						if not error["check"]:
							topoGeomErrors.append((error["intersection"], "Overlapped area"))
							errorCount += 1
						topoErrors.append(error["check"])
						checkSingleFileGeom.append(error["check"])
					if all(checkSingleFileGeom):
						self.browser.append("Nessun errore riscontrato")
					else:
						self.browser.append("Riscontrati %d errori di sovrapposizione" % errorCount)


		self.browser.append("<br />")
		self.browser.append("<b><i>[5/5]</i> - Controllo completato</b>")
		self.browser.append("<br />")

		mainCheck[5] = all(topoErrors)

		QApplication.processEvents()
		self.setRecord(5, 0, True) if mainCheck[5] else self.setRecord(5, 0, False)


		if pointGeomErrors or topoGeomErrors:
			pointErrors = ErrorShapefile(0, 32633)
			lineErrors = ErrorShapefile(1, 32633)
			polyErrors = ErrorShapefile(2, 32633)

			for feature in pointGeomErrors:
				pointErrors.addFeature(feature)

			for feature in topoGeomErrors:
				if feature[0].type() == 1:
					lineErrors.addFeature(feature)
				if feature[0].type() == 2:
					polyErrors.addFeature(feature)

			pointErrorsLayer = pointErrors.shapefile
			lineErrorsLayer = lineErrors.shapefile
			polyErrorsLayer = polyErrors.shapefile

			mapRegistry = QgsMapLayerRegistry.instance()
			mapRegistry.addMapLayer(pointErrorsLayer)
			mapRegistry.addMapLayer(lineErrorsLayer)
			mapRegistry.addMapLayer(polyErrorsLayer)


		msgBox = QMessageBox()
		msgBox.setWindowTitle("Completato")
		msgBox.setText("Procedura di validazione completata")
		msgBox.exec_()




		self.saveButton.setEnabled(True)
		self.closeButton.setEnabled(True)




		"""
		for col in range(len(ProgressUi._HEADERS)):
			record = 0
			for field in range(len(ProgressUi._INITVALUES[0])):

			newitem = QTableWidgetItem(record[col])
			if col == 3:
				font = newitem.font()
				font.setBold(True)
				newitem.setTextColor(QColor("darkGreen"))
				newitem.setFont(font)
		self.recordNum += 1
		self.table.insertRow(self.recordNum)
		"""






if __name__ == "__main__":
	app = QApplication(sys.argv)
	form = ProgressUi()
	form.show()
	app.exec_()
